import logging
import os
import random

import faiss
import numpy as np
import pandas as pd
import pyarrow.parquet as pq

LOGGER = logging.getLogger(__name__)

from autofaiss import build_index
from tests.unit.test_embeddings_iterators import build_test_collection_numpy, build_test_collection_parquet

logging.basicConfig(level=logging.DEBUG)

# hide py4j DEBUG from pyspark, otherwise, there will be too many useless debugging output
# https://stackoverflow.com/questions/37252527/how-to-hide-py4j-java-gatewayreceived-command-c-on-object-id-p0
logging.getLogger("py4j").setLevel(logging.ERROR)


def test_quantize(tmpdir):
    min_size = random.randint(1, 100)
    max_size = random.randint(min_size, 10240)
    dim = random.randint(1, 100)
    nb_files = random.randint(1, 5)

    tmp_dir, sizes, dim, expected_array = build_test_collection_numpy(
        tmpdir, min_size=min_size, max_size=max_size, dim=dim, nb_files=nb_files
    )

    output_numpy_index = os.path.join(tmpdir.strpath, "numpy_knn.index")
    output_numpy_index_infos = os.path.join(tmpdir.strpath, "numpy_knn_infos.json")

    build_index(
        embeddings=tmp_dir,
        file_format="npy",
        index_path=output_numpy_index,
        index_infos_path=output_numpy_index_infos,
        max_index_query_time_ms=10.0,
        max_index_memory_usage="1G",
        current_memory_available="2G",
    )
    output_numpy_index_faiss = faiss.read_index(output_numpy_index)
    assert output_numpy_index_faiss.ntotal == len(expected_array)

    tmp_dir, sizes, dim, expected_df = build_test_collection_parquet(
        tmpdir, min_size=min_size, max_size=max_size, dim=dim, nb_files=nb_files
    )

    index_path = os.path.join(tmpdir.strpath, "parquet_knn.index")
    index_infos_path = os.path.join(tmpdir.strpath, "infos.json")

    build_index(
        embeddings=tmp_dir,
        file_format="parquet",
        embedding_column_name="embedding",
        index_path=index_path,
        index_infos_path=index_infos_path,
        max_index_query_time_ms=10.0,
        max_index_memory_usage="1G",
        current_memory_available="2G",
    )
    output_parquet_index_faiss = faiss.read_index(index_path)
    assert output_parquet_index_faiss.ntotal == len(expected_df)


def test_quantize_with_ids(tmpdir):
    min_size = random.randint(1, 100)
    max_size = random.randint(min_size, 10240)
    dim = random.randint(1, 100)
    nb_files = random.randint(1, 5)

    tmp_dir, sizes, dim, expected_df = build_test_collection_parquet(
        tmpdir, min_size=min_size, max_size=max_size, dim=dim, nb_files=nb_files
    )

    index_path = os.path.join(tmpdir.strpath, "parquet_knn.index")
    index_infos_path = os.path.join(tmpdir.strpath, "infos.json")
    ids_path = os.path.join(tmpdir.strpath, "ids")

    build_index(
        embeddings=tmp_dir,
        file_format="parquet",
        embedding_column_name="embedding",
        index_path=index_path,
        index_infos_path=index_infos_path,
        ids_path=ids_path,
        max_index_query_time_ms=10.0,
        max_index_memory_usage="1G",
        current_memory_available="2G",
        id_columns=["id"],
    )
    output_parquet_index_faiss = faiss.read_index(index_path)
    output_parquet_ids = pq.read_table(ids_path).to_pandas()
    assert output_parquet_index_faiss.ntotal == len(expected_df)

    expected_df["i"] = np.arange(start=0, stop=len(expected_df))
    pd.testing.assert_frame_equal(
        output_parquet_ids.reset_index(drop=True), expected_df[["id", "i"]].reset_index(drop=True)
    )


def test_quantize_with_pyspark(tmpdir):
    min_size = random.randint(1, 100)
    max_size = random.randint(min_size, 10240)
    dim = random.randint(1, 100)
    nb_files = random.randint(1, 5)
    tmp_dir, _, _, expected_df = build_test_collection_parquet(
        tmpdir, min_size=min_size, max_size=max_size, dim=dim, nb_files=nb_files
    )
    index_parquet_path = os.path.join(tmpdir.strpath, "parquet_knn.index")
    output_parquet_index_infos = os.path.join(tmpdir.strpath, "infos.json")
    ids_path = os.path.join(tmpdir.strpath, "ids")
    temporary_indices_folder = os.path.join(tmpdir.strpath, "distributed_autofaiss_indices")
    build_index(
        embeddings=tmp_dir,
        distributed="pyspark",
        file_format="parquet",
        temporary_indices_folder=temporary_indices_folder,
        index_infos_path=output_parquet_index_infos,
        ids_path=ids_path,
        max_index_memory_usage="1G",
        current_memory_available="2G",
        id_columns=["id"],
        embedding_column_name="embedding",
        index_path=index_parquet_path,
    )
    output_parquet_index_faiss = faiss.read_index(index_parquet_path)
    output_parquet_ids = pq.read_table(ids_path).to_pandas()
    assert output_parquet_index_faiss.ntotal == len(expected_df)
    pd.testing.assert_frame_equal(output_parquet_ids.reset_index(drop=True), expected_df[["id"]].reset_index(drop=True))

    tmp_dir, _, _, expected_array = build_test_collection_numpy(
        tmpdir, min_size=min_size, max_size=max_size, dim=dim, nb_files=nb_files
    )
    output_numpy_index = os.path.join(tmpdir.strpath, "numpy_knn.index")
    output_numpy_index_infos = os.path.join(tmpdir.strpath, "numpy_knn_infos.json")

    build_index(
        embeddings=tmp_dir,
        distributed="pyspark",
        file_format="npy",
        temporary_indices_folder=temporary_indices_folder,
        index_infos_path=output_numpy_index_infos,
        max_index_memory_usage="1G",
        current_memory_available="2G",
        embedding_column_name="embedding",
        index_path=output_numpy_index,
    )

    output_numpy_index_faiss = faiss.read_index(output_numpy_index)
    assert output_numpy_index_faiss.ntotal == len(expected_array)


def test_quantize_with_multiple_inputs(tmpdir):
    min_size = random.randint(1, 100)
    max_size = random.randint(min_size, 10240)
    dim = random.randint(1, 100)
    nb_files = random.randint(1, 5)
    tmp_dir1, _, _, expected_df1 = build_test_collection_parquet(
        tmpdir, min_size=min_size, max_size=max_size, dim=dim, nb_files=nb_files, tmpdir_name="autofaiss_parquet1"
    )
    tmp_dir2, _, _, expected_df2 = build_test_collection_parquet(
        tmpdir, min_size=min_size, max_size=max_size, dim=dim, nb_files=nb_files, tmpdir_name="autofaiss_parquet2"
    )
    expected_df = pd.concat([expected_df1, expected_df2])
    index_parquet_path = os.path.join(tmpdir.strpath, "parquet_knn.index")
    output_parquet_index_infos = os.path.join(tmpdir.strpath, "infos.json")
    build_index(
        embeddings=[tmp_dir1, tmp_dir2],
        file_format="parquet",
        embedding_column_name="embedding",
        index_path=index_parquet_path,
        index_infos_path=output_parquet_index_infos,
        max_index_query_time_ms=10.0,
        max_index_memory_usage="1G",
        current_memory_available="2G",
    )
    output_parquet_index_faiss = faiss.read_index(index_parquet_path)
    assert output_parquet_index_faiss.ntotal == len(expected_df)



""" Functions that read efficiently files stored on disk """
import logging
from multiprocessing.pool import ThreadPool
from typing import Iterator, Optional, Tuple, List, Union
import re
from abc import ABC

import pandas as pd
import numpy as np
from tqdm import tqdm as tq
import fsspec
import os
import pyarrow.parquet as pq


LOGGER = logging.getLogger(__name__)


class AbstractArray:
    num_rows: int

    def get_rows(self, start: int, end: int) -> Tuple[np.ndarray, Optional[pd.DataFrame]]:
        pass


class AbstractMatrixReader(ABC):
    """Read a file and provide its shape, row count and ndarray. Behaves as a context manager"""

    def __init__(self, fs: fsspec.AbstractFileSystem, file_path: str):
        self.f = fs.open(file_path, "rb")

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.f.close()

    def get_row_count(self) -> int:
        pass

    def get_shape(self) -> Tuple[int, int]:
        pass

    def get_lazy_array(self) -> AbstractArray:
        pass

    def get_eager_array(self) -> AbstractArray:
        pass


class NumpyEagerNdArray(AbstractArray):
    def __init__(self, f: Union[str, fsspec.core.OpenFile]):
        self.n = np.load(f)
        self.shape = self.n.shape
        self.num_rows = self.shape[0]

    def get_rows(self, start: int, end: int) -> Tuple[np.ndarray, Optional[pd.DataFrame]]:
        return self.n[start:end, :], None


def read_numpy_header(f):
    f.seek(0)
    file_size = f.size if isinstance(f.size, int) else f.size()
    first_line = f.read(min(file_size, 300)).split(b"\n")[0]
    result = re.search(r"'shape': \(([0-9]+), ([0-9]+)\)", str(first_line))
    shape = (int(result.group(1)), int(result.group(2)))
    dtype = re.search(r"'descr': '([<f0-9]+)'", str(first_line)).group(1)
    end = len(first_line) + 1  # the first line content and the endline
    f.seek(0)
    return (shape, dtype, end)


class NumpyLazyNdArray(AbstractArray):
    """Reads a numpy file lazily"""

    def __init__(self, f: fsspec.spec.AbstractBufferedFile):
        self.f = f
        (self.shape, self.dtype, self.header_offset) = read_numpy_header(f)
        self.byteperitem = np.dtype(self.dtype).itemsize * self.shape[1]
        self.num_rows = self.shape[0]

    def get_rows(self, start: int, end: int) -> Tuple[np.ndarray, Optional[pd.DataFrame]]:
        length = end - start
        self.f.seek(self.header_offset + start * self.byteperitem)
        return (
            np.frombuffer(self.f.read(length * self.byteperitem), dtype=self.dtype).reshape((length, self.shape[1])),
            None,
        )


class NumpyMatrixReader(AbstractMatrixReader):
    """Read a numpy file and provide its shape, row count and ndarray. Behaves as a context manager"""

    def __init__(self, fs: fsspec.AbstractFileSystem, file_path: str, *_):
        super().__init__(fs, file_path)

    def get_shape(self) -> Tuple[int, int]:
        shape, _, _ = read_numpy_header(self.f)
        return shape

    def get_row_count(self) -> int:
        return self.get_shape()[0]

    def get_eager_array(self) -> NumpyEagerNdArray:
        return NumpyEagerNdArray(self.f)

    def get_lazy_array(self) -> NumpyLazyNdArray:
        return NumpyLazyNdArray(self.f)


class ParquetEagerNdArray(AbstractArray):
    def __init__(self, f, embedding_column_name: str, id_columns: Optional[List[str]] = None):
        emb_table = pq.read_table(f).to_pandas()
        embeddings_raw = emb_table[embedding_column_name].to_numpy()
        self.n = np.stack(embeddings_raw)
        self.ids = emb_table[id_columns] if id_columns else None
        self.num_rows = self.n.shape[0]

    def get_rows(self, start, end) -> Tuple[np.ndarray, Optional[pd.DataFrame]]:
        return self.n[start:end, :], self.ids.iloc[start:end] if self.ids else None


class ParquetLazyNdArray(AbstractArray):
    """Reads a parquet file lazily"""

    def __init__(self, f, embedding_column_name: str, id_columns: Optional[List[str]] = None):
        self.table = pq.read_table(f)
        self.num_rows = self.table.num_rows
        self.embedding_column_name = embedding_column_name
        self.id_columns = id_columns

    def get_rows(self, start, end) -> Tuple[np.ndarray, Optional[pd.DataFrame]]:
        table_slice = self.table.slice(start, end - start)
        embeddings_raw = table_slice[self.embedding_column_name].to_numpy()
        ids = table_slice.select(self.id_columns).to_pandas() if self.id_columns else None
        return np.stack(embeddings_raw), ids


class ParquetMatrixReader(AbstractMatrixReader):
    """Read a parquet file and provide its shape, row count and ndarray. Behaves as a context manager"""

    def __init__(self, fs, file_path, embedding_column_name, id_columns=None):
        super().__init__(fs, file_path)
        self.embedding_column_name = embedding_column_name
        self.id_columns = id_columns

    def get_shape(self) -> Tuple[int, int]:
        parquet_file = pq.ParquetFile(self.f, memory_map=True)
        num_rows = parquet_file.metadata.num_rows
        batches = parquet_file.iter_batches(batch_size=1, columns=[self.embedding_column_name])
        dimension = next(batches).to_pandas()[self.embedding_column_name].to_numpy()[0].shape[0]
        return (num_rows, dimension)

    def get_row_count(self) -> int:
        parquet_file = pq.ParquetFile(self.f, memory_map=True)
        return parquet_file.metadata.num_rows

    def get_eager_array(self) -> ParquetEagerNdArray:
        return ParquetEagerNdArray(self.f, self.embedding_column_name, self.id_columns)

    def get_lazy_array(self) -> ParquetLazyNdArray:
        return ParquetLazyNdArray(self.f, self.embedding_column_name, self.id_columns)


matrix_readers_registry = {"parquet": ParquetMatrixReader, "npy": NumpyMatrixReader}


def get_matrix_reader(file_format: str, fs: fsspec.AbstractFileSystem, file_path: str, *args) -> AbstractMatrixReader:
    cls = matrix_readers_registry[file_format]
    return cls(fs, file_path, *args)


def get_file_list(path: Union[str, List[str]], file_format: str) -> Tuple[fsspec.AbstractFileSystem, List[str]]:
    """
    Get the file system and all the file paths that matches `file_format` under the given `path`.
    The `path` could a single folder or multiple folders.

    :raises ValueError: if file system is inconsistent under different folders.
    """
    if isinstance(path, str):
        return _get_file_list(path, file_format)
    all_file_paths = []
    fs = None
    for p in path:
        cur_fs, file_paths = _get_file_list(p, file_format, sort_result=False)
        if fs is None:
            fs = cur_fs
        elif fs != cur_fs:
            raise ValueError(
                f"The file system in different folder are inconsistent.\n" f"Got one {fs} and the other {cur_fs}"
            )
        all_file_paths.extend(file_paths)
    all_file_paths.sort()
    return fs, all_file_paths


def make_path_absolute(path: str) -> str:
    fs, p = fsspec.core.url_to_fs(path)
    if fs.protocol == "file":
        return os.path.abspath(p)
    return path


def _get_file_list(
    path: str, file_format: str, sort_result: bool = True
) -> Tuple[fsspec.AbstractFileSystem, List[str]]:
    """Get the file system and all the file paths that matches `file_format` given a single path."""
    path = make_path_absolute(path)
    fs, path_in_fs = fsspec.core.url_to_fs(path)
    prefix = path[: path.index(path_in_fs)]
    glob_pattern = path.rstrip("/") + f"**/*.{file_format}"
    file_paths = fs.glob(glob_pattern)
    if sort_result:
        file_paths.sort()
    file_paths_with_prefix = [prefix + file_path for file_path in file_paths]
    return fs, file_paths_with_prefix


def read_first_file_shape(
    embeddings_path: Union[str, List[str]], file_format: str, embedding_column_name: Optional[str] = None
) -> Tuple[int, int]:
    """
    Read the shape of the first file in the embeddings directory.
    """
    fs, file_paths = get_file_list(embeddings_path, file_format)

    first_file_path = file_paths[0]
    return get_file_shape(first_file_path, file_format, embedding_column_name, fs)


def get_file_shape(
    file_path: str, file_format: str, embedding_column_name: Optional[str], fs: fsspec.AbstractFileSystem
):
    with get_matrix_reader(file_format, fs, file_path, embedding_column_name) as matrix_reader:
        return matrix_reader.get_shape()


def read_total_nb_vectors_and_dim(
    embeddings_path: Union[str, List[str]], file_format: str = "npy", embedding_column_name: str = "embeddings"
) -> Tuple[int, int]:
    """
        Get the count and dim of embeddings.
        Parameters
        ----------
        embeddings_path : str
            Path of the embedding in numpy or parquet format.
        file_format : str (default "npy")

        Returns
        -------
        (count, dim) : (int, int)
            count: total number of vectors in the dataset.
            dim: embedding dimension
        """
    fs, file_paths = get_file_list(embeddings_path, file_format)

    _, dim = get_file_shape(file_paths[0], file_format=file_format, embedding_column_name=embedding_column_name, fs=fs)

    def file_to_line_count(f):
        with get_matrix_reader(file_format, fs, f, embedding_column_name) as matrix_reader:
            return matrix_reader.get_row_count()

    count = 0
    i = 0
    with tq(total=len(file_paths)) as pbar:
        with ThreadPool(50) as p:
            for c in p.imap_unordered(file_to_line_count, file_paths):
                count += c
                i += 1
                pbar.update(1)

    return count, dim


def read_embeddings(
    embeddings_path: Union[str, List[str]],
    batch_size: Optional[int] = None,
    verbose: bool = True,
    file_format: str = "npy",
    embedding_column_name: str = "embeddings",
    id_columns: Optional[List[str]] = None,
) -> Iterator[Tuple[np.ndarray, Optional[pd.DataFrame]]]:
    """
    Iterate over embeddings arrays.
    It is possible to iterate over batchs of files and yield stacked embeddings arrays.

    The implementation adopted here is chosen for memory concern: it is very important
    for autofaiss to avoid using more memory than necessary.
    In particular, for the faiss training, a large embedding array is necessary.
    It is not possible to save it twice in memory.
    This implementation pre-allocate an array of size batch_size and keep it updated during the iteration over files.
    The maximum memory usage is batch_size * dim * 4

    Parameters
    ----------
    embeddings_path : str
        Path on local disk of the embedding in numpy format.
    batch_size : int (default None)
        Outputs a maximum of batch_size vectors, the default is the size of the first file
        This parameter is useful when working with many small files.
    file_format : str (default "npy")
        Format of the embedding files.
        npy or parquet
    verbose : bool
        Print detailed informations if set to True
    embedding_column_name: str
        If file_format="parquet" - the name of the column containing the embeddings
    id_columns: List[str] (default None)
        List of columns containing the ids of the vectors,
        these will be mapped to integer ids between 0 and total_nb_embeddings

    Returns
    -------
    embeddings_iterator : Iterator[Tuple[np.ndarray, Optional[pd.DataFrame]]]
        An iterator over batches of stacked embedding arrays
        and (optionally) a DataFrame containing the mapping between id columns
        and an index in [0, total_nb_embeddings-1] store in a column "i"
    """
    try:
        first_vector_count, dim = read_first_file_shape(
            embeddings_path, file_format, embedding_column_name=embedding_column_name
        )
    except StopIteration as err:
        raise Exception("no files to read from") from err

    if batch_size is None:
        batch_size = first_vector_count

    fs, file_paths = get_file_list(embeddings_path, file_format)
    embeddings_batch = None
    ids_batch = None
    nb_emb_in_batch = 0

    iterator = file_paths
    if verbose:
        iterator = tq(list(iterator))

    total_embeddings_processed = 0
    for file_path in iterator:
        with get_matrix_reader(file_format, fs, file_path, embedding_column_name, id_columns) as matrix_reader:
            array = matrix_reader.get_lazy_array()
            vec_size = array.num_rows
            current_emb_index = 0
            while True:
                left_in_emb = vec_size - current_emb_index
                remaining_to_add = max(batch_size - nb_emb_in_batch, 0)
                adding = min(remaining_to_add, left_in_emb)
                additional = max(left_in_emb - adding, 0)
                if embeddings_batch is None:
                    embeddings_batch = np.empty((batch_size, dim), "float32")
                current_embeddings, ids_df = array.get_rows(current_emb_index, (current_emb_index + adding))
                embeddings_batch[nb_emb_in_batch : (nb_emb_in_batch + adding), :] = current_embeddings

                if id_columns is not None:
                    if ids_batch is None:
                        ids_batch = np.empty((batch_size, len(id_columns)), dtype="object")
                    ids_batch[nb_emb_in_batch : (nb_emb_in_batch + adding), :] = ids_df.to_numpy()

                nb_emb_in_batch += adding
                current_emb_index += adding
                if nb_emb_in_batch == batch_size:
                    if id_columns is not None:
                        ids_batch_df = pd.DataFrame(ids_batch, columns=id_columns).infer_objects()
                        ids_batch_df["i"] = np.arange(
                            start=total_embeddings_processed, stop=total_embeddings_processed + nb_emb_in_batch
                        )
                    else:
                        ids_batch_df = None

                    yield embeddings_batch, ids_batch_df
                    total_embeddings_processed += nb_emb_in_batch
                    nb_emb_in_batch = 0
                    embeddings_batch = None
                    ids_batch = None
                    ids_batch_df = None
                if additional == 0:
                    break

    if nb_emb_in_batch > 0 and embeddings_batch is not None:
        if id_columns is not None:
            ids_batch_df = pd.DataFrame(
                ids_batch[:nb_emb_in_batch, :], columns=id_columns  # type: ignore
            ).infer_objects()
            ids_batch_df["i"] = np.arange(
                start=total_embeddings_processed, stop=total_embeddings_processed + nb_emb_in_batch
            )
        else:
            ids_batch_df = None
        total_embeddings_processed += nb_emb_in_batch
        yield embeddings_batch[:nb_emb_in_batch], ids_batch_df



"""
Building the index with pyspark.
"""
import math
import multiprocessing
import os
from functools import partial
from multiprocessing.pool import ThreadPool
from tempfile import TemporaryDirectory
from typing import List, Optional, Iterator, Tuple, Callable, Any, Union

import faiss
import fsspec
import numpy as np
import pandas as pd
from fsspec import get_filesystem_class
from fsspec.implementations.hdfs import PyArrowHDFS
from tqdm import tqdm

from autofaiss.indices.index_utils import _get_index_from_bytes, _get_bytes_from_index
from autofaiss.readers.embeddings_iterators import get_file_list, get_matrix_reader, make_path_absolute
from autofaiss.utils.cast import cast_memory_to_bytes
from autofaiss.utils.decorators import Timeit


def _yield_embeddings_batch(
    embeddings_paths: List[str],
    chunk_sizes: List[int],
    file_system: fsspec.AbstractFileSystem,
    start: int,
    end: int,
    embedding_column_name: str,
    file_format: str,
    id_columns: Optional[List[str]] = None,
):
    """Lazy load a batch of embeddings."""
    if isinstance(file_system, PyArrowHDFS):
        file_system = get_filesystem_class("hdfs")()
    cur_start = cur_end = 0
    for chunk_size, file_path in zip(chunk_sizes, embeddings_paths):
        cur_end += chunk_size
        if cur_end < start:
            cur_start += chunk_size
            continue
        slice_start = max(0, start - cur_start)
        slice_end = min(chunk_size, end - cur_start)
        with get_matrix_reader(file_format, file_system, file_path, embedding_column_name, id_columns) as matrix_reader:
            yield matrix_reader.get_lazy_array().get_rows(start=slice_start, end=slice_end)
        if cur_end > end:
            break
        cur_start += chunk_size


def _generate_small_index_file_name(batch_id: int) -> str:
    return f"index_{batch_id}"


def _save_small_index(index: faiss.Index, batch_id: int, small_indices_folder: str) -> None:
    """Save index for one batch."""
    fs = _get_file_system(small_indices_folder)
    fs.mkdirs(small_indices_folder, exist_ok=True)
    small_index_filename = _generate_small_index_file_name(batch_id)
    with fsspec.open(small_index_filename, "wb").open() as f:
        faiss.write_index(index, faiss.PyCallbackIOWriter(f.write))
    dest_filepath = os.path.join(small_indices_folder, small_index_filename)
    fs.put(small_index_filename, dest_filepath)


def _add_index(
    start: int,
    end: int,
    broadcast_trained_index_bytes,
    file_system: fsspec.AbstractFileSystem,
    chunk_sizes: List[int],
    embeddings_file_paths: List[str],
    embedding_column_name: str,
    batch_id: int,
    small_indices_folder: str,
    file_format: str,
    id_columns: Optional[List[str]] = None,
    num_cores: Optional[int] = None,
    embedding_ids_df_handler: Optional[Callable[[pd.DataFrame, int], Any]] = None,
):
    """
    Add a batch of embeddings on trained index and save this index.

    Parameters
    ----------
    start: int
        Start position of this batch
    end: int
        End position of this batch
    broadcast_trained_index_bytes: pyspark.Broadcast
        Trained yet empty index
    chunk_sizes: List[int]
        A list of number of vectors in each embedding files
    embeddings_file_paths: List[str]
        A list of embeddings file paths
    embedding_column_name: str
        Embeddings column name for parquet ; default "embedding"
    batch_id: int
        Batch id
    small_indices_folder: str
        The folder where we save all the small indices
    num_cores: int
        Number of CPU cores (not Vcores)
    file_format: str
        Embedding file format, "npy" or "parquet"
    id_columns: Optional[List[str]]
        Names of the columns containing the Ids of the vectors, only used when file_format is "parquet"
    embedding_ids_df_handler: Optional[Callable[[pd.DataFrame, int], Any]]
        The function that handles the embeddings Ids when id_columns is given
    """
    total_chunk_sizes = sum(chunk_sizes)
    if end > total_chunk_sizes:
        end = total_chunk_sizes
    if len(chunk_sizes) != len(embeddings_file_paths):
        raise ValueError(
            f"The length of chunk_sizes should be equal to the number of embeddings_file_paths.\n"
            f"Got len(chunk_sizes)={len(chunk_sizes)},"
            f"len(embeddings_file_paths)={len(embeddings_file_paths)}.\n"
            f"chunk_sizes={chunk_sizes}\n"
            f"embeddings_file_paths={embeddings_file_paths}"
        )
    batch_vectors_gen, batch_ids_gen = zip(
        *_yield_embeddings_batch(
            embeddings_paths=embeddings_file_paths,
            chunk_sizes=chunk_sizes,
            file_system=file_system,
            start=start,
            end=end,
            embedding_column_name=embedding_column_name,
            id_columns=id_columns,
            file_format=file_format,
        )
    )

    embeddings_to_add = np.vstack(batch_vectors_gen).astype(np.float32)  # faiss requires float32 type

    if embedding_ids_df_handler is not None:
        embedding_ids_df_handler(pd.concat(batch_ids_gen), batch_id)

    if num_cores is None:
        num_cores = multiprocessing.cpu_count()

    faiss.omp_set_num_threads(num_cores)

    # load empty trained index
    empty_index = _get_index_from_bytes(broadcast_trained_index_bytes.value)

    empty_index.add(embeddings_to_add)

    del embeddings_to_add

    _save_small_index(index=empty_index, small_indices_folder=small_indices_folder, batch_id=batch_id)


def _get_pyspark_active_session():
    """Reproduce SparkSession.getActiveSession() available since pyspark 3.0."""
    import pyspark  # pylint: disable=import-outside-toplevel

    # pylint: disable=protected-access
    ss: Optional[pyspark.sql.SparkSession] = pyspark.sql.SparkSession._instantiatedSession  # mypy: ignore
    if ss is None:
        print("No pyspark session found, creating a new one!")
        ss = (
            pyspark.sql.SparkSession.builder.config("spark.driver.memory", "16G")
            .master("local[1]")
            .appName("Distributed autofaiss")
            .config("spark.submit.deployMode", "client")
            .getOrCreate()
        )
    return ss


def _batch_loader(batch_size: int, nb_batches: int) -> Iterator[Tuple[int, int, int]]:
    """Yield [batch id, batch start position, batch end position]"""
    for batch_id in range(nb_batches):
        start = batch_size * batch_id
        end = batch_size * (batch_id + 1)
        yield batch_id, start, end


def _download_one(src_dst_path: Tuple[str, str], fs: fsspec.AbstractFileSystem):
    src_path, dst_path = src_dst_path
    fs.get(src_path, dst_path)


def _parallel_download_indices_from_remote(
    fs: fsspec.AbstractFileSystem, indices_file_paths: List[str], dst_folder: str
):
    """Download small indices in parallel."""
    if len(indices_file_paths) == 0:
        return
    os.makedirs(dst_folder, exist_ok=True)
    dst_paths = [os.path.join(dst_folder, os.path.split(p)[-1]) for p in indices_file_paths]
    src_dest_paths = zip(indices_file_paths, dst_paths)
    with ThreadPool(min(16, len(indices_file_paths))) as pool:
        for _ in pool.imap_unordered(partial(_download_one, fs=fs), src_dest_paths):
            pass


def _merge_index(small_indices_folder: str, max_size_on_disk: str = "100GB") -> faiss.Index:
    """Merge all the indices in `small_indices_folder` into single one return the merged index."""
    fs = _get_file_system(small_indices_folder)
    small_indices_files = fs.ls(small_indices_folder, detail=False)
    if len(small_indices_files) == 0:
        raise ValueError(f"No small index is saved in {small_indices_folder}")

    def _get_index_from_file(filepath: str) -> faiss.Index:
        with open(filepath, "rb") as f:
            index_bytes = f.read()
        return _get_index_from_bytes(index_bytes)

    def _merge_from_local(merged: Optional[faiss.Index] = None) -> faiss.Index:
        local_file_paths = [
            os.path.join(local_indices_folder, filename) for filename in os.listdir(local_indices_folder)
        ]
        if merged is None:
            merged = _get_index_from_file(local_file_paths[0])
            start_index = 1
        else:
            start_index = 0

        for rest_index_file in tqdm(local_file_paths[start_index:]):
            index = _get_index_from_file(rest_index_file)
            faiss.merge_into(merged, index, shift_ids=True)
        return merged

    # estimate index size by taking the first index
    first_index_file = small_indices_files[0]
    first_index_size = fs.size(first_index_file)
    max_sizes_in_bytes = cast_memory_to_bytes(max_size_on_disk)
    nb_files_each_time = math.floor(max_sizes_in_bytes / first_index_size)
    merged_index = None
    n = len(small_indices_files)
    nb_iterations = max(math.ceil(n / nb_files_each_time), 1)
    with Timeit("-> Merging small indices", indent=4):
        with tqdm(total=nb_iterations) as pbar:
            for i in range(nb_iterations):
                to_downloads = small_indices_files[i * nb_files_each_time : min(n, (i + 1) * nb_files_each_time)]
                with TemporaryDirectory() as local_indices_folder:
                    _parallel_download_indices_from_remote(
                        fs=fs, indices_file_paths=to_downloads, dst_folder=local_indices_folder
                    )
                    merged_index = _merge_from_local(merged_index)
                pbar.update(1)
    return merged_index


def _get_chunk_sizes(
    embed_paths: List[str], embedding_column_name: str, id_columns: Optional[List[str]], file_format: str
) -> List[int]:
    """Get chunk sizes from a list of embeddings files."""

    def _get_parquet_row_count(embed_path: str) -> int:
        with get_matrix_reader(
            file_format, _get_file_system(embed_path), embed_path, embedding_column_name, id_columns
        ) as matrix_reader:
            return matrix_reader.get_row_count()

    chunk_sizes = []
    # use ThreadPool instead of multiprocessing.Pool to avoid having problem in _get_parquet_row_count
    with ThreadPool(16) as pool:
        for row_count in pool.imap(_get_parquet_row_count, embed_paths):
            chunk_sizes.append(row_count)
    return chunk_sizes


def run(
    faiss_index: faiss.Index,
    embeddings_path: Union[str, List[str]],
    batch_size: int,
    embedding_column_name: str = "embedding",
    num_cores_per_executor: Optional[int] = None,
    temporary_indices_folder="hdfs://root/tmp/distributed_autofaiss_indices",
    file_format: str = "npy",
    id_columns: Optional[List[str]] = None,
    embedding_ids_df_handler: Optional[Callable[[pd.DataFrame, int], Any]] = None,
) -> faiss.Index:
    """
    Create indices by pyspark.

    Parameters
    ----------
    faiss_index: faiss.Index
        Trained faiss index
    embeddings_path: str
        Embeddings folder
    batch_size: int
        Number of vectors handled per worker
    embedding_column_name: str
        Embeddings column name for parquet; default "embedding"
    num_cores_per_executor: int
        Number of CPU cores per executor
    temporary_indices_folder: str
        Folder to save the temporary small indices
    file_format: str
        Embeddings file format; default "npy"
        "npy" or "parquet"
    id_columns: Optional[List[str]]
        Names of the columns containing the Ids of the vectors, only used when file_format is "parquet"
    embedding_ids_df_handler: Optional[Callable[[pd.DataFrame, int], Any]]
        The function that handles the embeddings Ids when id_columns is given
    """
    if isinstance(embeddings_path, str):
        embeddings_path = make_path_absolute(embeddings_path)
    else:
        embeddings_path = list(map(make_path_absolute, embeddings_path))
    temporary_indices_folder = make_path_absolute(temporary_indices_folder)

    ss = _get_pyspark_active_session()
    # broadcast the index bytes
    trained_index_bytes = _get_bytes_from_index(faiss_index)
    broadcast_trained_index_bytes = ss.sparkContext.broadcast(trained_index_bytes)
    _, embed_paths = get_file_list(path=embeddings_path, file_format=file_format)
    chunk_sizes = _get_chunk_sizes(embed_paths, embedding_column_name, id_columns=id_columns, file_format=file_format)
    nb_vectors = sum(chunk_sizes)
    nb_batches = math.ceil(nb_vectors / batch_size)  # use math.ceil to make sure that we cover every vector
    batches = _batch_loader(batch_size=batch_size, nb_batches=nb_batches)
    rdd = ss.sparkContext.parallelize(batches, nb_batches)
    file_system = _get_file_system(embed_paths[0])
    with Timeit("-> Adding indices", indent=2):
        rdd.foreach(
            lambda x: _add_index(
                batch_id=x[0],
                start=x[1],
                end=x[2],
                broadcast_trained_index_bytes=broadcast_trained_index_bytes,
                embeddings_file_paths=embed_paths,
                file_system=file_system,
                chunk_sizes=chunk_sizes,
                embedding_column_name=embedding_column_name,
                id_columns=id_columns,
                small_indices_folder=temporary_indices_folder,
                num_cores=num_cores_per_executor,
                embedding_ids_df_handler=embedding_ids_df_handler,
                file_format=file_format,
            )
        )

    with Timeit("-> Merging indices", indent=2):
        merged_index = _merge_index(temporary_indices_folder)

        fs = _get_file_system(temporary_indices_folder)
        fs.rm(temporary_indices_folder, recursive=True)

    return merged_index


def _get_file_system(path: str) -> fsspec.AbstractFileSystem:
    return fsspec.core.url_to_fs(path)[0]
